/*
Think script for Cheese
	by Steel Titanium

Handles following Cream and reacting to what's happening to Cream.
*/

//Freeslot the required stuff
freeslot(
	"SPR_CHEZ",
	"S_CHEESEIDLE",
	"S_CHEESEFLY",
	"S_CHEESEHAPPY1",
	"S_CHEESEHAPPY2",
	"S_CHEESEHAPPY3",
	"S_CHEESEHAPPY4",
	"S_CHEESESAD1",
	"S_CHEESESAD2",
	"S_CHEESESAD3",
	"S_CHEESESAD4",
	"MT_CHEESE"
)

COM_AddCommand("togglecheese", function(player, onoff)
	if not player.cheesedisabled then
		CONS_Printf(player, "\nCream will keep Cheese in her pocket.")
		player.cheesedisabled = true
	else
		CONS_Printf(player, "\nCheese will provide air support.")
		player.cheesedisabled = false
	end
end)

mobjinfo[MT_CHEESE] = {
	doomednum = -1,
	spawnstate = S_CHEESEIDLE,
	spawnhealth = 1,
	painchance = 200,
	speed = 3,
	radius = 16*FRACUNIT,
	height = 32*FRACUNIT,
	dispoffset = 0,
	mass = 100,
	damage = 0,
	reactiontime = 32,
	activesound = sfx_None,
	flags = MF_NOGRAVITY|MF_FLOAT|MF_SCENERY|MF_NOCLIPHEIGHT|MF_DONTENCOREMAP,
	raisestate = S_NULL
}

states[S_CHEESEIDLE] = {
	sprite = SPR_CHEZ,
	frame = A|FF_ANIMATE,
	tics = -1,
	var1 = 1,
	var2 = 2,
	nextstate = S_CHEESEIDLE
}

states[S_CHEESEFLY] = {
	sprite = SPR_CHEZ,
	frame = C|FF_ANIMATE,
	tics = -1,
	var1 = 1,
	var2 = 2,
	nextstate = S_CHEESEFLY
}

states[S_CHEESESAD1] = {
	sprite = SPR_CHEZ,
	frame = H,
	tics = 5,
	var1 = 0,
	var2 = 0,
	nextstate = S_CHEESESAD2
}

states[S_CHEESESAD2] = {
	sprite = SPR_CHEZ,
	frame = I,
	tics = 3,
	var1 = 0,
	var2 = 0,
	nextstate = S_CHEESESAD3
}

states[S_CHEESESAD3] = {
	sprite = SPR_CHEZ,
	frame = J,
	tics = 6,
	var1 = 0,
	var2 = 0,
	nextstate = S_CHEESESAD4
}

states[S_CHEESESAD4] = {
	sprite = SPR_CHEZ,
	frame = I,
	tics = 3,
	var1 = 0,
	var2 = 0,
	nextstate = S_CHEESESAD1
}

states[S_CHEESEHAPPY1] = {
	sprite = SPR_CHEZ,
	frame = E,
	tics = 8,
	nextstate = S_CHEESEHAPPY2
}

states[S_CHEESEHAPPY2] = {
	sprite = SPR_CHEZ,
	frame = F,
	tics = 4,
	nextstate = S_CHEESEHAPPY3
}

states[S_CHEESEHAPPY3] = {
	sprite = SPR_CHEZ,
	frame = G,
	tics = 8,
	nextstate = S_CHEESEHAPPY4
}

states[S_CHEESEHAPPY4] = {
	sprite = SPR_CHEZ,
	frame = F,
	tics = 4,
	nextstate = S_CHEESEHAPPY1
}

addHook("MobjThinker", function(mo)
	if mo and mo.skin == "cream" and mo.player and not mo.player.cheesedisabled then //Skin check
		local player = mo.player
		if not player.mo.cheese
			player.mo.cheese = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_CHEESE)
			player.mo.cheese.angle = player.mo.angle
			player.mo.cheese.flags2 = player.mo.flags2
			player.mo.cheese.scale = player.mo.scale
			player.mo.cheese.target = player.mo
		end
	elseif (mo and mo.skin ~= "cream" and mo.cheese and mo.cheese.valid) or (mo.player.cheesedisabled) then // Not using Cream but still has Cheese with them.
		P_RemoveMobj(mo.cheese)
		mo.cheese = nil
	end
end, MT_PLAYER)

addHook("MobjThinker", function(mo)
	-- Thinker for Cheese
	if mo and mo.valid then
		if mo.target and mo.target.valid and mo.target.player and mo.target.player.valid then
			local tg = mo.target
			local totz

			-- Colorize Cheese if the player has invincibility, grow etc.
			mo.color = tg.color
			mo.colorized = tg.colorized

			-- Check for gravity flip
			if (mo.eflags & MFE_VERTICALFLIP)
				totz = (tg.z) - (tg.height)
			else
				totz = (tg.z) + (tg.height)
			end

			local mangle = FixedAngle(((leveltime * 5) % 360) << FRACBITS)
			mo.scale = tg.scale
			mo.angle = tg.angle
			mo.eflags = tg.eflags
			mo.flags2 = tg.flags2

			if (tg.player.exiting > 0 and K_IsPlayerLosing(tg.player) == true) or tg.player.kartstuff[k_spinouttimer] > 0 or tg.health == 0
				if not (mo.state >= S_CHEESESAD1 and mo.state <= S_CHEESESAD4)
					mo.state = S_CHEESESAD1
				end
				mo.z = totz
				P_TeleportMove(mo, tg.x, tg.y, totz)
				mo.angle = tg.angle
			end

			if tg.health > 0 and tg.player.kartstuff[k_spinouttimer] == 0 then
				if tg.player.speed > 0 or (tg.momx + tg.momy + tg.momz) > 0 then // If player is moving
					if mo.prevtpos == nil then
						mo.prevtpos = {}
					end

					if not (mo.state == S_CHEESEFLY)
						mo.state = S_CHEESEFLY
					end
					local trailby = 2

					if #mo.prevtpos >= trailby then
						local e = mo.prevtpos[trailby]
						local etotz

						if (mo.eflags & MFE_VERTICALFLIP)
							etotz = e.z - (tg.height) - (sin(FixedAngle(leveltime*5*FRACUNIT))*8)
						else
							etotz = e.z + (tg.height) + (sin(FixedAngle(leveltime*5*FRACUNIT))*8)
						end

						mo.prevtpos[trailby] = nil
						P_TeleportMove(mo, e.x, e.y, etotz)
					end

					table.insert(mo.prevtpos, 1, {x = tg.x, y = tg.y, z = tg.z})
				else -- If player is idle
					P_TeleportMove(mo, tg.x, tg.y, totz)

					if tg.player.exiting > 0 and K_IsPlayerLosing(tg.player) == false
						if not (mo.state >= S_CHEESEHAPPY1 and mo.state <= S_CHEESEHAPPY4)
							mo.state = S_CHEESEHAPPY1
						end
					elseif tg.player.exiting > 0 and K_IsPlayerLosing(tg.player) == true
						if not (mo.state >= S_CHEESESAD1 and mo.state <= S_CHEESESAD4)
							mo.state = S_CHEESESAD1
						end
					else
						if not (mo.state == S_CHEESEIDLE)
							mo.state = S_CHEESEIDLE
						end
					end

					mo.angle = tg.angle
					mo.momx = sin(mangle) * (tg.radius>>FRACBITS)
					mo.momy = cos(mangle) * (tg.radius>>FRACBITS)
				end
			end
		end
	end

	if mo and (not mo.target or not mo.target.valid) or (not mo.target.player or not mo.target.valid) then
		P_RemoveMobj(mo)
    end

end, MT_CHEESE)