-- FallBlame by Indev for Sunflower's Garden
-- Replicate HP MOD's fall event, except at all time with a few custom additions
local cv_fb_debug = CV_RegisterVar({
    name = "fb_debug",
    defaultvalue = "Off",
    possiblevalue = CV_OnOff,
})

local FALL_BLAME_TIME = 2*TICRATE

local function set(t)
    local ret = {}

    for i = 1, #t do ret[t[i]] = true end

    return ret
end

local CHECK_COLLISION = set {
    MT_ORBINAUT,
    MT_ORBINAUT_SHIELD,
    MT_JAWZ,
    MT_JAWZ_DUD,
    MT_JAWZ_SHIELD,
}

local function generateHFForPlayer(p)
	if p and p.valid and p.mo and p.mo.valid then
		return {{p.name, p.mo.hf_overridenamecolor or p.skincolor}, {skins[p.mo.skin].facemmap, p.skincolor}, p}
	else
		return {{}, {}}
	end
end

local function dprint(str, ...)
    if not cv_fb_debug.value then return end

    print(string.format("[fb_debug] "..str, ...))
end

local function setFallBlame(victim, attacker)
    dprint("Set %s as fallblame for %s", attacker.name, victim.name)

    victim.fallblame_timer = FALL_BLAME_TIME
    victim.fallblame = attacker
end

local function clearFallBlame(victim)
    dprint("Clear fallblame for %s", victim.name)

    victim.fallblame_timer = nil
    victim.fallblame = nil
end

local function sendHitfeedEvent(victim, attacker)
    if hpmod and hpmod.running then return end

    if attacker and not attacker.valid then
        attacker = nil
    else
        if attacker and victim ~= attacker then
            if attacker.mo and attacker.mo.valid then
                K_PlayHitEmSound(attacker.mo)
            end
        else
            -- You played yourself idiot
            attacker = nil
        end
    end

    local vname = victim and victim.name or "[REDACTED]"
    local aname = attacker and attacker.name or "[REDACTED]"

    dprint("%s fall off the map because of %s", vname, aname)

    if hitfeed then
        if not HF_animIcon["K_HMFAL"] then
            --print("no fall icon registered registering")
            HF_animIcon["K_HMFAL"] = {4, 6}
            --print("registered fall icon")
        end

        HF_SendHitMessage(generateHFForPlayer(attacker), generateHFForPlayer(victim), "K_HMFAL")
    end

    clearFallBlame(victim)
end

local function collide(pmo, mo)
    if pmo.z > (mo.z + mo.height) or mo.z > (pmo.z + pmo.height) then return end

    local p1 = pmo.player
    local p2

    if mo.player then
        local p2 = mo.player

        if (p1.powers[pw_flashing] and not p1.kartstuff[k_spinouttimer]) or (p2.powers[pw_flashing] and not p2.kartstuff[k_spinouttimer]) then return end

        setFallBlame(p2, p1)
    elseif CHECK_COLLISION[mo.type] then
        p2 = mo.target and mo.target.player
    end

    if p2 and (p2 ~= p1) then
        setFallBlame(p1, p2)
    end
end

addHook("MobjCollide", collide, MT_PLAYER)
addHook("MobjMoveCollide", collide, MT_PLAYER)

addHook("PlayerSpin", function(player, _, inflictor)
    if inflictor and inflictor.player then
        setFallBlame(player, inflictor.player)
    end
end)

-- Rare cases of player being squished in air
addHook("PlayerSquish", function(player, _, inflictor)
    if inflictor and inflictor.player then
        setFallBlame(player, inflictor.player)
    end
end)

addHook("PlayerExplode", function(player, _, inflictor)
    if inflictor and inflictor.player then
        setFallBlame(player, inflictor.player)
    end
end)

addHook("MobjDeath", function(pmo, mo, pmo2)
    if mo and mo.type == MT_SINK then return end
    if pmo.player.pflags & PF_TIMEOVER then return end

    local player = pmo.player

    sendHitfeedEvent(player, player.fallblame)
end, MT_PLAYER)

addHook("MobjThinker", function(pmo)
    -- pmo can be killed in previously called hooks i think
    if not (pmo.valid and pmo.player) then return end

    local player = pmo.player

    if player.fallblame_timer
        and P_IsObjectOnGround(pmo)
        and not player.kartstuff[k_spinouttimer]
        and not player.kartstuff[k_wipeoutslow] then

        player.fallblame_timer = $ - 1
    end

    if player.fallblame_timer == 0 then
        clearFallBlame(player)
    end
end, MT_PLAYER)
