--[[
	changelog

	v0.1.1.1 (LET'S GO)
		Changed method for predicting variables to maintain userdata compatibility with source functions

	v0.1.1
		Fixed bonk playing when players wouldn't collide in the first place
		Otherwise revised logic for checking for goomba bonks to be more accurate

	v0.1
		initial
]]

freeslot("sfx_ccobnk")

-- cache regularly used global constants and functions
local G_BattleGametype = G_BattleGametype

local k_squishedtimer = k_squishedtimer
local k_hyudorotimer = k_hyudorotimer
local k_justbumped = k_justbumped
local k_bumper = k_bumper
local pw_flashing = pw_flashing
local k_invincibilitytimer = k_invincibilitytimer
local k_growshrinktimer = k_growshrinktimer
local k_comebacktimer = k_comebacktimer
local k_comebackmode = k_comebackmode
local TICRATE = TICRATE
local k_spinouttimer = k_spinouttimer
local PST_LIVE = PST_LIVE
local k_respawn = k_respawn

-- things that should be constants in source
local KART_BATTLE_COMEBACK_MODE_ITEM = 1

local sfx_ccobnk = sfx_ccobnk

S_sfx[sfx_ccobnk] = {
	flags = SF_NOINTERRUPT
}

local function deepCopy(originalValue)
	if type(originalValue) ~= "table" then
		if type(originalValue) == "userdata" then
			print("\130CAUTION: Attempt to deep-copy userdata, which can't be deep-copied!")
		end
		return originalValue
	end

	local tableCopy = {}
	for tableKey, tableValue in pairs(originalValue) do
		-- DO NOT DEEP COPY KEY; a table copy hashes differently!
		tableCopy[tableKey] = deepCopy(tableValue)
	end
	setmetatable(tableCopy, deepCopy(getmetatable(originalValue)))

	return tableCopy
end

local function copyToTableLike(sourceTable, targetTableLike)
	for sourceKey, sourceValue in pairs(sourceTable) do
		if type(sourceValue) == "table" then
			local targetValue = targetTableLike[sourceKey]
            local targetValueType = type(targetValue)
			if targetValueType == "table" or targetValueType == "userdata" then
				copyToTableLike(sourceValue, targetValue)
			else
				targetTableLike[sourceKey] = sourceValue
			end
		else
			targetTableLike[sourceKey] = sourceValue
		end
	end
end

local function isMobjValidPlayer(mobj)
	-- spectating players have already been checked
	return mobj.player and mobj.player.valid
end

local function areMobjsCollidingVertically(mobjA, mobjB)
	return mobjA.z <= mobjB.z + mobjB.height and mobjA.z + mobjA.height >= mobjB.z
end

local function isPlayerSquished(player)
	return player.kartstuff[k_squishedtimer] > 0
end

local function isPlayerUsingHyudoro(player)
	return player.kartstuff[k_hyudorotimer] > 0
end

local function didPlayerRecentlyBumpSomething(player)
	return player.kartstuff[k_justbumped] > 0
end

local function canPlayerCrushOtherPlayer(crusher, other)
	return crusher.mo.scale > other.mo.scale + mapobjectscale / 8
end

local function doesPlayerHaveBumpers(player)
	return player.kartstuff[k_bumper] > 0
end

local function isMobjFallingOnMobjOnGround(collidingMobj, collidedMobj)
	return P_IsObjectOnGround(collidedMobj) and collidingMobj.momz < 0
end

local function isPlayerAboutToSquishOtherPlayer(squishingPlayerMobj, otherPlayerMobj)
	return squishingPlayerMobj.scale > otherPlayerMobj.scale + mapobjectscale / 8
end

local function isPlayerFlashing(player)
	return player.powers[pw_flashing] > 0
end

local function isPlayerInvincible(player)
	return player.kartstuff[k_invincibilitytimer] > 0
end

local function isPlayerGrown(player)
	return player.kartstuff[k_growshrinktimer] > 0 or
		-- HugeQuest interop
		player.hugequest and
			player.hugequest.huge > 0
end

local function isPlayerWaitingForCombeback(player)
	return player.kartstuff[k_comebacktimer] > 0
end

local function isPlayerRespawning(player)
	return player.kartstuff[k_respawn] > 0
end

local function isFlashingPlayerCollidable(player)
	return player.powers[pw_flashing] >= K_GetKartFlashing(player)
end

local function isBumperedCollidingWithBumperless(collidedPlayer, collidingPlayer)
	return doesPlayerHaveBumpers(collidedPlayer) and not doesPlayerHaveBumpers(collidingPlayer)
end

local function shouldPlayersCollide(collidingPlayer, collidedPlayer)
	return not (isPlayerSquished(collidedPlayer) or
			isPlayerUsingHyudoro(collidedPlayer) or
			didPlayerRecentlyBumpSomething(collidedPlayer) or
			canPlayerCrushOtherPlayer(collidedPlayer, collidingPlayer) or
			isPlayerSquished(collidingPlayer) or
			isPlayerUsingHyudoro(collidingPlayer) or
			didPlayerRecentlyBumpSomething(collidingPlayer) or
			canPlayerCrushOtherPlayer(collidingPlayer, collidedPlayer) or
			G_BattleGametype() and
				(isBumperedCollidingWithBumperless(collidedPlayer, collidingPlayer) or
					isBumperedCollidingWithBumperless(collidingPlayer, collidedPlayer))
		)
end

local function shouldPlayerBeSquished(targetPlayer)
	return not (
			targetPlayer.mo.health <= 0 and
			isPlayerFlashing(targetPlayer) or
			isPlayerSquished(targetPlayer) or
			isPlayerInvincible(targetPlayer) or
			isPlayerGrown(targetPlayer) or
			isPlayerUsingHyudoro(targetPlayer) or
			G_BattleGametype() and
				(not doesPlayerHaveBumpers(targetPlayer) and
					isPlayerWaitingForCombeback(targetPlayer) or
				targetKartStuff[k_comebackmode] == KART_BATTLE_COMEBACK_MODE_ITEM)
		)
end

local function isPlayerAboutToDamageOtherPlayer(damagingPlayer, otherPlayer)
	return isPlayerInvincible(damagingPlayer) and not isPlayerInvincible(otherPlayer)
end

local function shouldPlayerBeDamaged(targetPlayer)
	return not (
			targetPlayer.mo.health <= 0 or
			targetPlayer.exiting or
			isPlayerInvincible(targetPlayer) or
			isPlayerGrown(targetPlayer) or
			isPlayerFlashing(targetPlayer)
		)
end

local function areMobjsPredictedToBeMovingTowardEachOther(mobjA, mobjB)
	local momentumDifferenceX = mobjA.momx - mobjB.momx
	local momentumDifferenceY = mobjA.momy - mobjB.momy

	local predictedDistanceX = mobjA.x + mobjB.momx * 3 - (mobjB.x + mobjA.momx * 3)
	local predictedDistanceY = mobjA.y + mobjB.momy * 3 - (mobjB.y + mobjA.momy * 3)
	if predictedDistanceX == 0 and predictedDistanceY == 0 then
		-- no distance to calculate angle properly
		return false
	end

	local predictedDistance = FixedHypot(predictedDistanceX, predictedDistanceY)
	local normalizedPredictedDistanceX = FixedDiv(predictedDistanceX, predictedDistance)
	local normalizedPredictedDistanceY = FixedDiv(predictedDistanceY, predictedDistance)

	predictedDistanceX = FixedMul(mobjA.radius + mobjB.radius, normalizedPredictedDistanceX)
	predictedDistanceY = FixedMul(mobjA.radius + mobjB.radius, normalizedPredictedDistanceY)

	if momentumDifferenceX == 0 and momentumDifferenceY == 0 then
		momentumDifferenceX = -normalizedPredictedDistanceX
		momentumDifferenceY = -normalizedPredictedDistanceY
	end

	local momentumDifferenceLength = FixedHypot(momentumDifferenceX, momentumDifferenceY)
	if momentumDifferenceLength < 25 * mapobjectscale then
		local normalizedMomentumDifferenceX = FixedDiv(momentumDifferenceX, momentumDifferenceLength)
		local normalizedMomentumDifferenceY = FixedDiv(momentumDifferenceY, momentumDifferenceLength)
		momentumDifferenceX = FixedMul(25 * mapobjectscale, normalizedMomentumDifferenceX)
		momentumDifferenceY = FixedMul(25 * mapobjectscale, normalizedMomentumDifferenceY)
	end

	local momentumDotProduct = FixedMul(momentumDifferenceX, predictedDistanceX) +
		FixedMul(momentumDifferenceY, predictedDistanceY)

	return momentumDotProduct < 0
end

local function shouldPlayersBounce(bouncingPlayerMobj, trouncedPlayerMobj)
	local bouncingPlayer = bouncingPlayerMobj.player
	local trouncedPlayer = trouncedPlayerMobj.player
	return not (
			bouncingPlayer.playerstate ~= PST_LIVE or
			trouncedPlayer.playerstate ~= PST_LIVE or
			isPlayerRespawning(bouncingPlayer) or
			isPlayerRespawning(trouncedPlayer)
		) and
		(not isPlayerFlashing(bouncingPlayer) or
			isFlashingPlayerCollidable(bouncingPlayer)) and
		(not isPlayerFlashing(trouncedPlayer) or
			isFlashingPlayerCollidable(trouncedPlayer)) and
		not (
			didPlayerRecentlyBumpSomething(bouncingPlayer) or
			didPlayerRecentlyBumpSomething(trouncedPlayer)
		) and
		areMobjsPredictedToBeMovingTowardEachOther(bouncingPlayerMobj, trouncedPlayerMobj)
end

local function predictGoombaBonkVariables(collidingPlayerMobj, collidedMobj)
	-- no, we can't just set it to the references, because they're mutable
	local originalVariables = nil

	-- based on p_map.c:1459
	if isMobjValidPlayer(collidingPlayerMobj) and
		isMobjValidPlayer(collidedMobj) and
		areMobjsCollidingVertically(collidingPlayerMobj, collidedMobj)
	then
		originalVariables = {
			collidedPlayer = {
				kartstuff = {
					[k_squishedtimer] = collidedMobj.player.kartstuff[k_squishedtimer]
				},
				powers = {
					[pw_flashing] = collidedMobj.player.powers[pw_flashing]
				}
			},
			collidingPlayer = {
				kartstuff = {
					[k_squishedtimer] = collidingPlayerMobj.player.kartstuff[k_squishedtimer]
				},
				powers = {
					[pw_flashing] = collidingPlayerMobj.player.powers[pw_flashing]
				}
			}
		}
		local predictedVariables = deepCopy(originalVariables)

		local collidingPlayer = collidingPlayerMobj.player
		local collidedPlayer = collidedMobj.player
		if isPlayerAboutToSquishOtherPlayer(collidingPlayerMobj, collidedMobj) then
			-- K_SquishPlayer will target collided player; this has lua hooks and will need interop
			if shouldPlayerBeSquished(collidedPlayer) then
				predictedVariables.collidedPlayer.kartstuff[k_squishedtimer] = TICRATE
				predictedVariables.collidedPlayer.powers[pw_flashing] = K_GetKartFlashing(collidedPlayer)
			end
		elseif isPlayerAboutToSquishOtherPlayer(collidingPlayerMobj, collidedMobj) then
			-- K_SquishPlayer will target colliding player; this has lua hooks and will need interop
			if shouldPlayerBeSquished(collidingPlayer) then
				predictedVariables.collidingPlayer.kartstuff[k_squishedtimer] = TICRATE
				predictedVariables.collidingPlayer.powers[pw_flashing] = K_GetKartFlashing(collidingPlayer)
			end
		elseif isPlayerAboutToDamageOtherPlayer(collidingPlayer, collidedPlayer) then
			-- P_DamageMobj will target collided player; this has lua hooks and will need interop
			if shouldPlayerBeDamaged(collidedPlayer) then
				predictedVariables.collidedPlayer.powers[pw_flashing] = K_GetKartFlashing(collidedPlayer)
			end
		elseif isPlayerAboutToDamageOtherPlayer(collidedPlayer, collidingPlayer) then
			-- P_DamageMobj will target colliding player; this has lua hooks and will need interop
			if shouldPlayerBeDamaged(collidingPlayer) then
				predictedVariables.collidingPlayer.powers[pw_flashing] = K_GetKartFlashing(collidingPlayer)
			end
		end

		-- change the players
		copyToTableLike(predictedVariables.collidedPlayer, collidedPlayer)
		copyToTableLike(predictedVariables.collidingPlayer, collidingPlayer)
	end


	return originalVariables
end

local function isCollisionPlayerGoombaBounce(collidingPlayerMobj, collidedMobj)
	local result = false

	-- some mechanics haven't run on this frame yet, so relevant variables must be predicted
	local originalVariables = predictGoombaBonkVariables(collidingPlayerMobj, collidedMobj)

	if collidedMobj.health ~= 0 and
		isMobjValidPlayer(collidedMobj) and
		areMobjsCollidingVertically(collidingPlayerMobj, collidedMobj) and
		shouldPlayersCollide(collidingPlayerMobj.player, collidedMobj.player)
	then
		if isMobjFallingOnMobjOnGround(collidingPlayerMobj, collidedMobj) then
			result = shouldPlayersBounce(collidingPlayerMobj, collidedMobj)
		end
		if isMobjFallingOnMobjOnGround(collidedMobj, collidingPlayerMobj) then
			result = shouldPlayersBounce(collidedMobj, collidingPlayerMobj)
		end
	end

	-- restore modified player variables
	-- ...wait, is anything in here?
	if originalVariables then
		copyToTableLike(originalVariables.collidedPlayer, collidedMobj.player)
		copyToTableLike(originalVariables.collidingPlayer, collidingPlayerMobj.player)
	end

	return result
end

local function checkGoombaBonk(collidingMobj, collidedMobj)
	-- replicate conditions leading up to goomba bounce in PIT_CheckThing
	-- mobj validity has already been checked
	if isCollisionPlayerGoombaBounce(collidingMobj, collidedMobj) then
		S_StartSoundAtVolume(collidedMobj, sfx_ccobnk, 255)
	end
end

addHook("MobjMoveCollide", checkGoombaBonk, MT_PLAYER)
