freeslot("S_HYUSPY", "SPR_HYSM")

states[S_HYUSPY] = {
    sprite = SPR_HYSM,
    frame = A|FF_ANIMATE,
    var1 = 3,
    var2 = 7,
    tics = 21,
    nextstate = S_HYUSPY,
}

local k_itemtype = k_itemtype
local k_itemamount = k_itemamount
local KITEM_HYUDORO = KITEM_HYUDORO

local MAXHYUDISPLAY = TICRATE/2
local MAXHYUOFFSET = 8*FRACUNIT
local MAXHYUMOBJOFFSET = 32*FRACUNIT
local HYUMOBJBASEOFFSET = 32*FRACUNIT

local cv_hyuspy = CV_RegisterVar {
    name = "hyuspy",
    defaultvalue = "Mobj",
    possiblevalue = {
        Off = 0,
        HUD = 1,
        Mobj = 2,
    },
}

local cv_hyuspy_mobj_transparency = CV_RegisterVar {
    name = "hyuspy_mobj_transparency",
    defaultvalue = 4,
    possiblevalue = {
        MIN = 0,
        MAX = 9,
    },
}

local function countHyudoros()
    local count = 0

    for player in players.iterate do
        if player.kartstuff[k_itemtype] == KITEM_HYUDORO then
            count = count + 1
        end
    end

    return count
end

local function rescale(value, oldmin, oldmax, newmin, newmax)
    return newmin + FixedMul(FixedDiv(value - oldmin, oldmax - oldmin), newmax - newmin)
end

-- Returns value from 0 to 10, where 0 is no transparency, 1-9 - translucency,
-- and 10 - invisible
local function getTransparency(minval, hyudisplay)
    return rescale(hyudisplay<<FRACBITS, MAXHYUDISPLAY<<FRACBITS, 0, minval<<FRACBITS, 10<<FRACBITS)>>FRACBITS
end

local function getOffset(hyudisplay)
    local t = FixedDiv(hyudisplay<<FRACBITS, MAXHYUDISPLAY<<FRACBITS)

    t = FixedSqrt(t)

    return rescale(t, 0, FRACUNIT, MAXHYUOFFSET, 0)
end

local function getMobjOffset(hyudisplay)
    local t = FixedDiv(hyudisplay<<FRACBITS, MAXHYUDISPLAY<<FRACBITS)

    t = FixedSqrt(t)

    return rescale(t, 0, FRACUNIT, MAXHYUMOBJOFFSET, 0)
end

local function hyuspypetThink(mo, player)
    local dp = false
    for i = 0, 3 do -- Should be #displayplayers-1, not 3, but fixdisplayplayers.lua breaks it
        if player == displayplayers[i] then
            dp = true
            break
        end
    end
    
    local trans = getTransparency(cv_hyuspy_mobj_transparency.value, player.hyudisplay or 0)
    
    if not (cv_hyuspy.value == 2 and dp and trans < 10) then
        mo.flags2 = mo.flags2 | MF2_DONTDRAW
    else
        mo.flags2 = mo.flags2 & ~MF2_DONTDRAW
    end
    
    local pmo = player.mo
    
    if pmo then
        if mo.flags2 & MF2_DONTDRAW then
            if pmo then
                P_SetOrigin(mo, pmo.x, pmo.y, pmo.z+pmo.height*2)
            end
        else
            local addangle = (5*leveltime/3)*ANG1
            local offset = getMobjOffset(player.hyudisplay)
            local xoff = FixedMul(cos(pmo.angle - ANGLE_90), offset) + FixedMul(cos(addangle), HYUMOBJBASEOFFSET)
            local yoff = FixedMul(sin(pmo.angle - ANGLE_90), offset) + FixedMul(sin(addangle), HYUMOBJBASEOFFSET)
            local zoff = 5*pmo.height/4
            
            xoff = FixedMul(xoff, mapobjectscale)
            yoff = FixedMul(yoff, mapobjectscale)
            zoff = FixedMul(zoff, mapobjectscale)
            
            P_MoveOrigin(mo, pmo.x + xoff, pmo.y + yoff, pmo.z + zoff)
            mo.frame = (mo.frame & ~FF_TRANSMASK) | trans<<FF_TRANSSHIFT
            mo.angle = addangle + ANGLE_90
        end
    end
end

addHook("ThinkFrame", function()
    local hyudoros = countHyudoros()

    for player in players.iterate do
        if player.spectator or player.exiting then
            player.hyudisplay = nil
            
            if player.hyuspypet and player.hyuspypet.valid then
                P_RemoveMobj(player.hyuspypet)
            end
            continue
        end
        
        local pmo = player.mo
        
        if (pmo and pmo.valid) and not (player.hyuspypet and player.hyuspypet.valid) then
            local hyuspypet = P_SpawnMobj(pmo.x, pmo.y, pmo.z, MT_THOK)
            
            hyuspypet.state = S_HYUSPY
            
            player.hyuspypet = hyuspypet
        end
        
        hyuspypetThink(player.hyuspypet, player)
        
        player.hyudisplay = player.hyudisplay or 0

        if player.kartstuff[k_stolentimer] > 0 then
            -- If player's item got stolen, hide hyudoro immideately
            player.hyudisplay = 0
        elseif hyudoros == 0                                                       -- No hyudoros
            or (hyudoros == 1 and player.kartstuff[k_itemtype] == KITEM_HYUDORO)   -- We are the one holding the only hyudoro
            or player.kartstuff[k_itemamount] == 0 or player.kartstuff[k_itemheld] ~= 0 -- We don't have any items to steal
            or player.kartstuff[k_itemblink] ~= 0 then                             -- Item is blinking, so not stealable

            -- Slowly fade hyudoro out
            player.hyudisplay = max(player.hyudisplay - 1, 0)
        else
            -- Hyudoro fades in and stares at your item
            player.hyudisplay = min(player.hyudisplay + 1, MAXHYUDISPLAY)
        end
    end
end)

-- Adjusts coords defined in 320x200 resolution for player in splitscreen
local function getSplitScreenCoords(p, x, y)
    if spliscreen == 0 then
        return x, y
    elseif splitscreen == 1 then
        return x, y/2 + p.splitscreenindex*100
    else
        return x/2 + (p.splitscreenindex%2)*160, y/2 + (p.splitscreenindex/2)*100
    end
end

local cv_itemoffx, cv_itemoffy
local function lookupCvars()
    -- Specific for sungarden, since there are item hud offsets in our version
    -- of xitemlib. There are also hud_item_xoffset and hud_item_yoffset in
    -- saturn build, but xitem will use them if it detects them. If this script
    -- used somewhere outside of sungarden feel free to replace hud_xitem_* with
    -- hud_item_*, or remove completely. They are not necessary.

    if cv_itemoffx then return end -- Found already

    cv_itemoffx = CV_FindVar("hud_xitem_xoffset")
    cv_itemoffy = CV_FindVar("hud_xitem_yoffset")
end

addHook("MapLoad", lookupCvars)
addHook("NetVars", lookupCvars)

-- Probably it should take splitscreen into account...
local function getItemPosition()
    local offx = (cv_itemoffx and cv_itemoffx.value or 0)
    local offy = (cv_itemoffy and cv_itemoffy.value or 0)

    return 5 + offx, 5 + offy
end

local hyu
hud.add(function(v, p)
    if hyu == nil then hyu = v.cachePatch("K_ISHYUD") end
    if p.hyudisplay == nil or cv_hyuspy.value ~= 1 then return end

    local trans = getTransparency(v.localTransFlag()>>V_ALPHASHIFT, p.hyudisplay)

    if trans == 10 then return end

    local flags = V_SNAPTOLEFT|V_SNAPTOTOP

    if trans ~= 0 then flags = flags | trans<<V_ALPHASHIFT end

    local x, y = getItemPosition()
    x = x + 64
    y = y + 24

    if splitscreen > 1 then
        x = x - 4
        y = y - 4
    end

    x, y = getSplitScreenCoords(p, x, y)

    v.drawScaled(x<<FRACBITS + getOffset(p.hyudisplay), y<<FRACBITS, FRACUNIT, hyu, flags)
end)
