-- Keeppoints by Indev for Sunflower's Garden
-- Keeps... points of a user for specific period of time

local points = {
    storage = {},
}

local cv_minkeeppoints = CV_RegisterVar {
    name = "minkeeppoints",
    defaultvalue = "20",
    PossibleValue = CV_Unsigned,
    flags = CV_NETVAR,
}

local cv_keeppointstime = CV_RegisterVar {
    name = "keeppointstime",
    defaultvalue = TICRATE*60*60,
    PossibleValue = CV_Unsigned,
    flags = CV_NETVAR,
}

local cv_debuglog = CV_RegisterVar {
    name = "keeppoints_debug_log",
    defaultvalue = "Off",
    PossibleValue = CV_OnOff,
}

local function log(fmt, ...)
    assert(type(fmt) == "string")

    if cv_debuglog.value then
        print("[keeppoints] "..string.format(fmt, ...))
    end
end

local function fmt_time(tics)
    local seconds = (tics / TICRATE) % 60
    local minutes = tics / (TICRATE*60) % 60
    local hours = tics / (TICRATE*60*60)

    return string.format("%d:%02d:%02d", hours, minutes, seconds)
end

function points:gc()
    local to_remove = {}

    for name, stored in pairs(self.storage) do
        if stored.score < cv_minkeeppoints.value or stored.tics <= 0 then
            table.insert(to_remove, name)
        end
    end

    for _, name in ipairs(to_remove) do
        self.storage[name] = nil

        log("Remove stored points for player '%s'", name)
    end
end

function points:think()
    for player in players.iterate do
        player.score = max(player.score, player.battlekeepscore or 0)
        player.battlekeepscore = player.score
    
        local stored = self.storage[player.name]

        if stored then
            if stored.score > player.score then
                if G_BattleGametype() then
                    player.battlekeepscore = stored.score
                else
                    player.score = stored.score
                end
                
                self.storage[player.name] = nil
                log("Restore %d points for player '%s'", stored.score, player.name)
            end
        end
    end

    local to_delete = {}

    for k, stored in pairs(self.storage) do
        stored.tics = stored.tics - 1
    end

    self:gc()
end

function points:store(player)
    local score = max(player.score, player.battlekeepscore or 0)

    if score < cv_minkeeppoints.value or not cv_keeppointstime.value then return end

    self.storage[player.name] = {
        score = score,
        tics = cv_keeppointstime.value,
    }

    log("Store %d points for player '%s'", score, player.name)
end

addHook("ThinkFrame", function() points:think() end)
addHook("IntermissionThinker", function() points:think() end)
addHook("VoteThinker", function() points:think() end)

addHook("PlayerQuit", function(player)
    points:store(player)
end)

addHook("NetVars", function(net)
    points.storage = net(points.storage)
end)

COM_AddCommand("showstoredpoints", function(p)
    for name, stored in pairs(points.storage) do
        print(string.format("Player name: '%s', stored points: %d, time left: %s", name, stored.score, fmt_time(stored.tics)))
    end
end)
