// Race Rule Limiter by TheOneGoofAli.
// This script was mostly written way before MapUtils and SkInfo, so this may
// look a lot rougher than either of the two. Edits have taken place since then,
// but the majority of the script is still the same.

if not RRLimit then rawset(_G, "RRLimit", {}) end
RRLimit.name = "RRLimit"
RRLimit.fullname = "Race Rule Limiter"
RRLimit.ver = "Kicking Debut (1.01b)"
RRLimit.author = "TheOneGoofAli"
RRLimit.starttimeoffset = 6*TICRATE + (3*TICRATE/4)

// The race rule thing.
local RRLimitEnabled = CV_RegisterVar({
	name = "RRLimitActive",
	defaultvalue = 0,
	flags = CV_NETVAR | CV_MODIFIED | CV_CALL,
	possiblevalue = CV_OnOff,
	func = function(cvar)
		local kt = CV_FindVar("kicktime")
		// Kicks are temporary bans as of 1.6, so...
		if cvar.value == 1 and kt.value != 0 then
			CV_Set(kt, 0)
			print("\x82"..RRLimit.name.."\x80"..": Set 'kicktime' to 0.")
		//else
		//	CV_Set(kt, kt.defaultvalue)
		end
	end
})

// Limit set.
local RRMaxCount = CV_RegisterVar({
	name = "RRLimitMax",
	defaultvalue = 5,
	flags = CV_NETVAR,
	possiblevalue = {MIN = 1, MAX = 9999}
})

// Initialize variables to use.
local function InitializeStats(p)
	if p.rrl then return end
	local rrl = {
		racecomplete = false,
		coursecount = 0,
		hasreachedlimit = false,
		limitwarning = false,
		markedforremoval = false,
		countproc = false
	}
	p.rrl = rrl
end

local InVoteScreen, InIntermission = false, false

// Changing the course count on a player.
// Add (key = 0), set (key = 1). To subtract, have amount be negative.
local function SetCourseCount(p, key, amount)
	key, amount = $ or 0, $ or 1
	if p and p.valid and p.rrl then
		if key == 0 then
			p.rrl.coursecount = $ + amount
		elseif key == 1 then
			p.rrl.coursecount = amount
		end
	end
end

// Map start stuff.
addHook("MapLoad", function()
	InVoteScreen, InIntermission = false, false
	for p in players.iterate()
		if p.rrl then
			if p.rrl.racecomplete then p.rrl.racecomplete = false end
			p.rrl.countproc = false // Reset race join mark.
		end
	end
end)

// Type of action to take.
local RRLimitAction = CV_RegisterVar({
	name = "RRLimitAction",
	defaultvalue = 0,
	flags = CV_NETVAR,
	possiblevalue = {warn = 0, kick = 1}
})

// If we kick, when do we kick?
local RRLimitKickMode = CV_RegisterVar({
	name = "RRLimitKickWhen",
	defaultvalue = 0,
	flags = CV_NETVAR,
	possiblevalue = {vote = 0, inter = 1, voteinter = 2, map = 3, asap = 4}
})

local function IsEligibleKME(mode)
	if RRLimitKickMode.value == 4 // Always.
	or (RRLimitKickMode.value == 2 and (mode == "vote" or mode == "inter")) // Both voting and intermission.
	or (RRLimitKickMode.value == 0 and mode == "vote") // Voting screen only.
	or (RRLimitKickMode.value == 1 and mode == "inter") // Intermission only.
	or (RRLimitKickMode.value == 3 and mode == "player") // In map. Not recommended.
		then return true
	else return false end
end

// Kick code moved outside of main hook code. Please don't kick the server.
local function ExecuteKick(p, reason)
	if p == server or players[0] == p then
		error("RRLimit will not kick the host player/node!")
	end
	local RRLConsole, textos = RRLimit.name + ": ", "Performing automated kick of " + p.name + "..."
	print("\x85"..RRLConsole.."\x80"..textos)
	local command = "kick \""..p.name.."\" Automated RRLimit kick."
	COM_BufInsertText(server, command)
end
// Function exposed.
RRLimit.ExecKick = ExecuteKick

local function CheckAndMark(p)
	local text, text1
	if p and p.valid and p.rrl.hasreachedlimit then
		if RRLimitAction.value == 0 and not p.rrl.limitwarning then // Warn.
			text = " reached the course limit."
			chatprint("\x82"..p.name.."\x80"..text, true)
			p.rrl.limitwarning = true
			p.rrl.hasreachedlimit = true
		elseif RRLimitAction.value == 1 and not p.rrl.markedforremoval then // Kick everyone.
			if IsPlayerAdmin(p) or p == server then
				if p == server or players[0] == p then
					if not p.rrl.limitwarning then
						text = " reached the course limit. Nothing will happen to the host player."
						print("\x82"..p.name.."\x80"..text)
						p.rrl.limitwarning = true
					end
					p.rrl.hasreachedlimit = true
				else
					text = " reached the course limit and will be kicked,"
					text1 = " regardless of administrative status."
					print("\x82"..p.name.."\x80"..text.."\x85"..text1)
					p.rrl.markedforremoval = true
					p.rrl.hasreachedlimit = true
				end
			else
				text = " reached the course limit and will be kicked."
				print("\x82"..p.name.."\x80"..text)
				p.rrl.markedforremoval = true
				p.rrl.hasreachedlimit = true
			end
		end
	end
end

// The player think 1.
addHook("PlayerThink", function(p)
	if RRLimitEnabled.value == 0 then return end
	if p and p.valid then
		if not p.rrl then
			InitializeStats(p)
			return
		end
		if leveltime >= 20*TICRATE + RRLimit.starttimeoffset and not p.spectator then p.rrl.countproc = true end // Player is racing or has joined the race in time.
		if p.exiting > 0 and not p.rrl.racecomplete then p.rrl.racecomplete = true end // Has player finished?
		if p.rrl.coursecount < RRMaxCount.value and (p.rrl.markedforremoval or p.rrl.hasreachedlimit or p.rrl.limitwarning) then
			// Set variables to false if they are true, but the course limit is higher than race count.
			p.rrl.hasreachedlimit = false
			p.rrl.markedforremoval = false
			p.rrl.limitwarning = false
		end
	end
end)

// Players leaving? Nullify everything.
addHook("PlayerQuit", function(p)
	p.rrl = nil
end)

// Did someone at least finish when exiting the map? Otherwise, assume
// we have used 'exitlevel' to exit a potentially dangerous map.
local function MapFinished(p)
	if p.rrl.racecomplete then return true else return false end
end

// Intermission stuff.
addHook("IntermissionThinker", function()
	if RRLimitEnabled.value == 1 and not InIntermission then
		local naturalexit = false
		for p in players.iterate()
			if MapFinished(p) then naturalexit = true end
		end
		if naturalexit then // If someone has successfully finished the map, count everyone up.
			for p in players.iterate()
				if p.rrl.coursecount >= 0 and p.rrl.coursecount < RRMaxCount.value then
					if p.rrl.countproc then SetCourseCount(p, 0, 1) end
					CONS_Printf(p, "\x82"..RRLimit.name.."\x80"..": Your course count is " .. p.rrl.coursecount .. ".")
				end
				if p.rrl.coursecount >= RRMaxCount.value and not p.rrl.hasreachedlimit then
					p.rrl.hasreachedlimit = true
					p.rrl.limitwarning = true
					CONS_Printf(p, "\x82"..RRLimit.name.."\x80"..": You have reached the course limit of "..RRMaxCount.value..". Please rotate out of the session.")
					CheckAndMark(p)
				end
				if p.rrl.hasreachedlimit and p.rrl.markedforremoval and p.rrl.coursecount >= RRMaxCount.value
				and IsEligibleKME("inter") and not p.rrl.targeted then
					p.rrl.targeted = true
					ExecuteKick(p)
				end
			end
		end
		InIntermission = true
	end
	InVoteScreen = false
end)

// Vote screen stuff.
addHook("VoteThinker", function()
	if RRLimitEnabled.value == 1 and not InVoteScreen then
		for p in players.iterate()
			if p and p.valid and p.rrl then
				if p.rrl.coursecount >= RRMaxCount.value and not p.rrl.hasreachedlimit then
					p.rrl.hasreachedlimit = true
					p.rrl.limitwarning = true
					CONS_Printf(p, "\x82"..RRLimit.name.."\x80"..": You have reached the course limit of "..RRMaxCount.value..". Please rotate out of the session.")
					CheckAndMark(p)
				end
				if p.rrl.coursecount >= RRMaxCount.value then
					if RRLimitAction.value == 1 and not (p == server or players[0] == p)
						p.rrl.markedforremoval = true
						p.rrl.hasreachedlimit = true
					end
				end
				// If the player had been marked for a kick and we kick during voting screens...
				if p.rrl.hasreachedlimit and p.rrl.markedforremoval and p.rrl.coursecount >= RRMaxCount.value
				and IsEligibleKME("vote") and not p.rrl.targeted then
					p.rrl.targeted = true
					ExecuteKick(p)
				end
			end
		end
		InVoteScreen = true
	end
	InIntermission = false
end)

local function NowLookAtThis(net)
	InVoteScreen, InIntermission = net($), net($)
end
addHook("NetVars", NowLookAtThis)

// Kick-related hook for Players.
addHook("PlayerThink", function(p)
	if RRLimitEnabled.value == 0 then return end
	if p and p.valid and p.rrl then
		if p.rrl.hasreachedlimit and p.rrl.markedforremoval and p.rrl.coursecount >= RRMaxCount.value
		and IsEligibleKME("player") and not p.rrl.targeted then
			p.rrl.targeted = true
			ExecuteKick(p)
		end
		if p.rrl.coursecount >= RRMaxCount.value then
			if RRLimitAction.value == 1 and not (p == server or players[0] == p)
				p.rrl.markedforremoval = true
				p.rrl.hasreachedlimit = true
			end
		end
	end
end)

local function CheckRaceCount(p)
	if p and p.valid and p.rrl then
		if not RRLimitEnabled.value then
			CONS_Printf(p, "\x82"..RRLimit.name.."\x80"..": The race rule is not in effect currently.")
			return
		end
		if not p.rrl.hasreachedlimit then
			CONS_Printf(p, "\x82"..RRLimit.name.."\x80"..": Your course count is "..p.rrl.coursecount)
		elseif p.rrl.coursecount >= RRMaxCount.value then
			CONS_Printf(p, "\x82"..RRLimit.name.."\x80"..": You have reached the course limit of "..RRMaxCount.value..".")
		else
			CONS_Printf(p, "\x82"..RRLimit.name.."\x80"..": You have reached the course limit of "..RRMaxCount.value..". Please rotate out of the session.")
		end
	else
		CONS_Printf(p, "\x82"..RRLimit.name.."\x80"..": The required variables have not been found - aborting with error...")
	end
end

local function GetPlayerByName(target, ttype)
	if type(target) == "string" then // Name check.
		target = target:lower()
		for tp in players.iterate()
			local tpname = tp.name:lower()
			if tpname:lower() == target then return tp end
		end
	else
		return consoleplayer // Fallback.
	end
end

local function AdminRaceCount(p, action, targettype, targetp, amount)
	targettype, targetp, amount = $ or "name", $ or p, $ or "nan"
	local thetarget
	if targettype == "name" then
		thetarget = GetPlayerByName(targetp)
	elseif targettype == "node" then
		thetarget = players[tonumber(targetp)] 
	end
	if thetarget == nil and not (targettype == "all" and action == "view") then
		CONS_Printf(p, "\x82"..RRLimit.name.."\x80"..": Invalid player name/node or an error occured, aborting...")
		return
	end
	if action == "view" then
		if targettype == "all" then
			for play in players.iterate do
				CONS_Printf(p, "\x82"..RRLimit.name.."\x80"..": "..play.name.."'s course count is "..play.rrl.coursecount)
			end
		elseif thetarget and thetarget.valid and thetarget.rrl then
			CONS_Printf(p, "\x82"..RRLimit.name.."\x80"..": "..thetarget.name.."'s course count is "..thetarget.rrl.coursecount)
		end
	elseif action == "set" then
		if amount == "nan" or tonumber(amount) == nil then
			CONS_Printf(p, "\x82"..RRLimit.name.."\x80"..": Amount not given or invalid number.")
			return
		end
		if thetarget and thetarget.valid and thetarget.rrl then
			SetCourseCount(thetarget, 1, tonumber(amount))
		end
	elseif action == "add"
		if amount == "nan" or tonumber(amount) == nil then
			CONS_Printf(p, "\x82"..RRLimit.name.."\x80"..": Amount not given or invalid number.")
			return
		end
		if thetarget and thetarget.valid and thetarget.rrl then
			SetCourseCount(thetarget, 0, tonumber(amount))
		end
	else
		CONS_Printf(p, "This command lets an administrator view/modify players' course counts.")
		local textthing = {
			{"rrlimitadmin view <name/node> <player name/node number>", " - view the player count."},
			{"rrlimitadmin set <name/node> <player name/node number> <amount>", " - set the player count to <amount>."},
			{"rrlimitadmin add <name/node> <player name/node number> <amount>", " - add <amount> the player count. <amount> can be negative to subtract instead."}
		}
		for i = 1, #textthing
			CONS_Printf(p, "\x85"..textthing[i][1].."\x80"..textthing[i][2])
		end
	end
end

// Heccing colors.
local function RRLimitMain(p)
	CONS_Printf(p, "\x82"..RRLimit.fullname.." "..RRLimit.ver.."\x80".." by ".."\x87"..RRLimit.author)
	CONS_Printf(p, "This script allows hosts to track and limit players on a course amount basis.")
	local textthing = {
		{"rrlimitcheck", " - see your own current course count."},
		{"rrlimitactive", " - administrative command, enable/disable this addon."},
		{"rrlimitmax", " - administrative CVar, set the course limit. Minimum is 1, maximum is 9999."},
		{"rrlimitadmin", " - administrative command, can view/edit anyone's course count."},
		{"rrlimitaction <action>", " - administrative CVar. 0 = warn, 1 = kick"},
		{"rrlimitkickwhen <mode>", " - administrative CVar, controls when to kick. 0 = voting screen, 1 = intermission, 2 = both voting and intermission, 3 = during a map, 4 = always"}
	}
	for i = 1, #textthing
		CONS_Printf(p, "\x82"..textthing[i][1].."\x80"..textthing[i][2])
	end
end
COM_AddCommand("rrlimit", RRLimitMain)
COM_AddCommand("rrlimitcheck", CheckRaceCount)
COM_AddCommand("rrlimitadmin", AdminRaceCount, COM_ADMIN)

// We loaded? We loaded.
print("\x82"..RRLimit.fullname.." "..RRLimit.ver.."\x80".." by ".."\x87"..RRLimit.author)